/*

DDS / controller for JUMA-TX1 CW Transmitter. 
OH2NLT Juha Niinikoski 23.11.2005

Processor is Microchip 16F88 providing 4k words Flash, 256 bytes RAM and 256 bytes EEPROM.
Project compiled under MPLAB7.20 with Hitec PICC 8.05PL2 c-compiler

This code can be used freely for private Ham Radio projects.
Code is not a sample of clean good c language programming.
Limited amount of microcontroller Flash memory have forced to use
not clean but code size effective programming methods.

Initial version 1.00 23.11.2005
This code is based on JUMA-RX1 v3.03 features
DDS is stopped during RX (TX signal == 0)
TX signal indicator to display 06.12.2005 v1.01

*/


//**********************
// Include header files
//**********************

#include 	<pic.h>
#include	"jumatx1.h"


//****************
// Hello messages
//*****************

// Note ! Display organization is 2x8 characters in one line

const char ver[] = {"OH2NLT 06.12.2005"};		// Version info
//const char lcd_ver1[] = {" OH7SV  "}; 
//const char lcd_ver2[] = {" OH2NLT "};


//******************
// Global variables
//******************

signed char tune_digit;				// which digit to tune
signed char old_tune_digit;
#define DEF_STEP 	2				// default / start up tune step

bit pwchange;						// encoder moved
signed char pwctr;					// encoder delta
unsigned char portb_copy;			// PORTB read data from irq routine. Update intervall 1ms
			
bit dmode;							// LCD display mode, only two alternatives
bit testmode;						// test mode indicator

int ms_timer;						// millisecond timer, decremented by irq routine, stop at 0

union								// Battery voltage from A/D converter
	{
	int battery;					// A/D counts = U/2
	unsigned char adcregs[2];
	} ad;

unsigned char smeter;				// S-meter data from A/D converter. Smax = 240mV

// VFO structure
	union							// VFO data
	{
	unsigned long vfos[2];			// VFO A = 0, B = 1
	unsigned char vfo_bytes[8];
	}bank1 vfo;

#define vfo_A			(vfo.vfos[0])	// define VFOs at structures
#define vfo_B			(vfo.vfos[1])

#define	dds_freq		(m1.m1l)	// define storage at large mul structures
//long dds_freq;						// frequency in use
long tune_step;						// tuning step

// tuning limits
#define TUNE_LOW_LIMIT	0			// test setup can go down to 0 Hz
#define TUNE_HIGH_LIMIT 0x6666666	// 8 MHz upper limit
#define TUNE_TESTH_LIMIT 132875550	// 9,9 MHz upper limit

// tuning steps
#define TUNE_STEP_LOW 	134			// 10 Hz
#define TUNE_STEP_MID	1342		// 100 Hz
//#define TUNE_STEP_MID	6711		// 500 Hz
#define TUNE_STEP_UP	13422		// 1 kHz
#define TUNE_STEP_HI	1342177		// 100 kHz

#define TX_SHIFT		9395		// 700 Hz, when TX subtract this

// user interface timers
#define SAVE_DELAY		2000		// button press time (ms) for save frequency

// some DDS constants
#define BITS28		268435456UL		// 2^28 constant
#define BITS32		4294967296UL	// 2^32 constant
#define DDC_CLK		20000000UL		// DDS Chip clock

#define DMULT		320000000UL		// display multiplier, 0,07450581Hz/bit * 2^32

#define START_FREQ	0x2F58E21		// factory default (empty EEPROM) frequency, 3.699.000 Hz
#define START_FREQ_A	0x2CED911	// VFO A 3.510.000 Hz
#define START_FREQ_B	0x59CAC80	// VFO B 7.015.000 Hz
#define MHz_cal		0xCCCCCC		// exact 1.000.000 Hz setup for calibration

#define FILTER_SEL	(0x3333333 >> 14)	// Filter select 0x3333333 = 4.000.000 Hz

// other
bit	aux_out;						// aux output bit. Transferred to output with DDS write.
bit	band;							// band filter select bit. 1 = 7MHz				

//*****************
// Include C files
//*****************

#include	"ad9833.c"
#include	"large_mul.c"
#include	"dcrxtim.c"
#include	"dcrxlcd.c"
#include 	"dcrxadc.c"


//******************
// Processor config
//******************

	__CONFIG(UNPROTECT & HS & WDTDIS & PWRTEN & BOREN & LVPDIS & MCLRDIS);	// 16F88 config
//	__CONFIG(2, FCMDIS & IESODIS);
// CONFIG2 does not work with 805PL2 compiler, patch manually

	__IDLOC(0007);				// Device ID


//******************
// Helper functions
//******************


// generate tone RA0 pin (battery voltage measurement)

void tone(int pitch, int len)		// pitch = 1/2 cycle in us units, len = ms
	{
	ms_timer = len;
	while(ms_timer != 0)
		{
		TONE_OUT = 1;
		delayu(pitch);
		TONE_OUT = 0;
		delayu(pitch);
		}
	}

// tone definitions

#if 0
#define LOW_TONE	2500	// 200 Hz
#define MID_TONE	833		// 600 Hz
#define HIGH_TONE	500		// 1 kHz
#endif

#define LOW_TONE	1911	// C 261,6 Hz
#define MID_TONE	1517	// E 329,6 Hz
#define HIGH_TONE	1276	// G 392 Hz
#define HIGH2_TONE	1012	// B 493,8 Hz

#define CHANGE_LEN	100		// step change beep length (ms)


#define MORSE_TONE	955		// 523,2 Hz
#define DI 0
#define DAA 1
#define UNIT 40				// morse unit length (ms)


// generate morse symbol

void beep(char l)
	{
	if(l == DI)
		tone(MORSE_TONE, UNIT);
	else
		tone(MORSE_TONE, (3*UNIT));
	delaym(UNIT);
	}

// say CW hello 

void morse_message(void)		// hello experiment
	{
	TONE_DIR = 0;				// make tone output active

	beep(DI);					// say JUMA
	beep(DAA);
	beep(DAA);
	beep(DAA);
	delaym(3*UNIT);
	beep(DI);
	beep(DI);
	beep(DAA);
	delaym(3*UNIT);
	beep(DAA);	
	beep(DAA);
	delaym(3*UNIT);
	beep(DI);
	beep(DAA);
	delaym(7*UNIT);				// word space

	beep(DAA);					// say TX1
	delaym(3*UNIT);
	beep(DAA);
	beep(DI);
	beep(DI);
	beep(DAA);
	delaym(3*UNIT);
	beep(DI);	
	beep(DAA);
	beep(DAA);
	beep(DAA);
	beep(DAA);
	delaym(3*UNIT);

	TONE_DIR = 1;				// return to input mode
//	delaym(7*UNIT);				// end delay
	}

void morse_roger(void)				// say roger
	{
	TONE_DIR = 0;				// make tone output active

	beep(DI);					// say R
	beep(DAA);
	beep(DI);

	TONE_DIR = 1;				// return to input mode
	}


// store & read frequency to/from internal EEPROM
// EEPROM[0...7] = frequency, EEPROM[8] = step
void save_freq(void)
	{
	char x;
	for(x=0; x<8; x++)
		{
		EEPROM_WRITE(x, (vfo.vfo_bytes[x]));		// save VFO data
		}
	EEPROM_WRITE(8, tune_digit);					// save tune digit
	}

void read_freq(void)
	{
	char x;
	for(x=0; x<8; x++)
		{
		vfo.vfo_bytes[x] = EEPROM_READ(x);			// restore VFO data
		}
	tune_digit = EEPROM_READ(8);					// restore tuning step
	}


// check rx_frequency (DDS tuning word) limits and
// force rx_freq into limits
// status = 0 if ok, 1 if low, 2 if high

char check_freq_limits(void)
	{
	if((long)dds_freq < TUNE_LOW_LIMIT)		// limitters
		{
		dds_freq = TUNE_LOW_LIMIT;			// low = 0
		return(1);
		}
	if(testmode)
		{
		if(dds_freq > TUNE_TESTH_LIMIT)
			{
			dds_freq = TUNE_TESTH_LIMIT;			// high = 9,9 MHz
			return(2);
			}
		}
	else
		{
		if(dds_freq > TUNE_HIGH_LIMIT)
			{
			dds_freq = TUNE_HIGH_LIMIT;			// high = 8 MHz
			return(2);
			}
		}

	return(0);
	}

// Generate beep tone according tune step 

const int tune_tone[5] = {0, LOW_TONE, MID_TONE, HIGH_TONE, HIGH2_TONE};	// 16F88 tune step tone indicators

void step_beep( void )
	{
	TONE_DIR = 0;								// make tone output active

	tone(tune_tone[tune_digit], CHANGE_LEN);	// select tone & beep

	TONE_DIR = 1;								// return to input mode
	} 




//*****************
// Main function
//*****************

void main(void)
{
/* Init IO Ports & software modules */

//Set up ADC and PORTs

	ANSEL = ANSEL_INIT;		// A/D converter pin select

	PORTA = PORTA_INIT;		// A/D & control signals
	PORTB = PORTB_INIT;		// LCD & SW port

	ADCON1 = ADCON1_INIT;	// PORTA = digital I/O during start

	TRISA = TRISA_INIT;		// set port direction bits
	TRISB = TRISB_INIT;

	aux_out = 0;			// set to defined state

	Init_timers();			// Init timers & interrupt system

	init_ad9833();			// init DDS
	tune_ad9833(0);			// turn it immediatelly off, no interference to RX during start up
	read_freq();			// read VFO start up values from EEPROM

	if(VFO_SEL == 1)		// select which VFO to put in use
		dds_freq = vfo_A;
	else
		dds_freq = vfo_B;

	if(check_freq_limits() != 0)			// check limits, empty EEPROM ?
		{
		if(VFO_SEL == 1)
			dds_freq = START_FREQ_A;		// if check fails use cold start definition
		else
			dds_freq = START_FREQ_B;

		vfo_A = START_FREQ_A;				// also synchronize VFOs
		vfo_B = START_FREQ_B;	
		tune_digit = DEF_STEP;
		}
	old_tune_digit = tune_digit;
//	tune_ad9833(dds_freq);	// set DDS

	lcdinit();				// Init LCD
	putst("JUMA-TX1");		// Show RX type & version
	lcdcom(SECOND_HALF);
	putst("  v1.01 ");
	delaym(200);			// wait for audio ready
	pwctr = 0;				// clear encoder
	morse_message();		// say audible hello
	lcdclr( );
	putst(" OH7SV  ");		// Tell who we are
	lcdcom(SECOND_HALF);
	putst(" OH2NLT ");
	delaym(1800);			// "optimized" message display time

	set_chgen();			// download soft characters
	lcdclr( );				// clear display


	if((portb_copy & BUTTON_1) == 0) // Oscillator test ?
		{
		putst("1 MHz");		// show calibration message
		tune_ad9833(MHz_cal);	// set 1.000.000 MHz
		vfo_A = MHz_cal;
		vfo_B = MHz_cal;	// set both VFOs to calibration freq
		delaym(10000);		// show message 10s
		testmode = 1;		// set test mode flag
		}

// Free running Main Control Loop
// Loop time: About 5ms when voltage display selected
// About 6,3ms when S-meter display selected
// 1ms button pressed.

	for(;;)
	{
// do A/D measurements

	measure_all();								// measure & store A/D channells


// addjust display

	if((portb_copy & BUTTON_1) == 0) 			// test encoder button function
		{
		lcdclr();
		pwctr = 0;
		pwchange = 0;							// clear encoder for setup function
		step_beep();							// entry generate beep, tell where we are
		ms_timer = SAVE_DELAY;					// arm save timer, ! note step_beep() use same timer !

		while((portb_copy & BUTTON_1) == 0)		// display & tune setup loop
			{
			tune_digit = tune_digit + pwctr;
			if(tune_digit<1)
				tune_digit = 1;					// check limits(1...4) & stay between them
			if(tune_digit>4)
				tune_digit = 4;

			pwctr = 0;

			lcdcom(FIRST_HALF);
			if(testmode)						// test mode display
				if(VFO_SEL == 1)				// select VFO
					putlonghex(vfo_A);			// show DDS control word
				else
					putlonghex(vfo_B);
			else
				{
				putst(" VFO ");					// show selected VFO
				if(VFO_SEL == 1)
					putst("A");
				else
					putst("B");	
				}

			lcdcom(SECOND_HALF);							// show which digit to tune
			putst("  Step ");
			puthexnible(tune_digit);

			if(tune_digit != old_tune_digit)	// generate beep if step changed
				{
				step_beep();					// generate beep
				old_tune_digit = tune_digit;	// save change
				}

			if(ms_timer == 0 && pwchange == 0)	// long press && encoder not moved
				{
				pwchange = 1;					// save only once per push
				lcdcom(FIRST_HALF);
				putst("Settings");
				lcdcom(SECOND_HALF);
				putst("  Saved ");				// show status
				save_freq();					// save frequency to EEPROM
				morse_roger();					// roger
				while((portb_copy & BUTTON_1) == 0) // wait here for button release
					;
				}

			}

		if(pwchange == 1)						// encoder turned ?
			{
			pwchange = 0;						//reset flag, do nothing more
			}
		else
			{
			dmode = !dmode;						// change display mode
			}
		}


// tune TX

	if(pwchange == 1)							// do encoder, addjust frequency
		{
		switch(tune_digit)						// 16F88 version of tuning steps
			{
			case 1: default:
			tune_step = TUNE_STEP_LOW;			// fine tuning
			break;

			case 2:
			tune_step = TUNE_STEP_MID;			//  mid range tuning
			break;

			case 3:
			tune_step = TUNE_STEP_UP;			//  mid/high range tuning
			break;

			case 4:
			tune_step = TUNE_STEP_HI;			//  coarse tuning
			break;
			}

		if(VFO_SEL == 1)						// select VFO to tune
			{
			vfo_A = vfo_A + pwctr * tune_step;	// add +/- steps
			}
		else
			{
			vfo_B = vfo_B + pwctr * tune_step;	// add +/- steps
			}
		
		pwchange = 0;							// clear encoder
		pwctr = 0;
		}


// select VFO

	if(VFO_SEL == 1)						// select VFO in use
		dds_freq = vfo_A;
	else
		dds_freq = vfo_B;

		check_freq_limits();				// check limits (dds_freq)

// evaluate band filter select

	band = 0;								// set band select flag. DDS driver writes flip-flop
	if( ((dds_freq >> 14) & 0x3FFF) > FILTER_SEL)	// use upper 14 bits only
		band = 1;							// 1 = 7MHz filter select, 0 = 3.5MHz select

// set DDS

	if(((portb_copy & TX) == 0) && (testmode == 0))	// frequency output only in TX or Test mode
		tune_ad9833(0);						// stop DDS
	else
		tune_ad9833(dds_freq);				// set DDS if TX on

// Format & Display selected page

	lcdcom(FIRST_HALF);						// select line one (first 8 chr) display
	if(dmode == 0)							// Display page 0
		{
		draw_s_meter(smeter);				// draw smeter graph, scale = 0...48 counts
		}
	else									// Display page 1
		{
		putvolt(ad.battery << 1);			// A/D count * 2 = voltage  12.00V
		if((portb_copy & TX) == 0)			// show Tx state indicator
			putst(" V ");
		else
			putst("Tx ");					// when Tx is on
		}

// calculate and display DDS frequency in use (dds_freq)
	
	lcdcom(SECOND_HALF);					// line two allways frequency display
//	m1.m1l = dds_freq;						// calculate display frequency from DDS setup word
	m2.m2l = DMULT;
	vlmul();								// very large multiply, 32*32 bits
	add_prod(3, 0x80);						// round, add 2^32/2 to product before truncation 26.01.2005
	putfreq(prod.prodlng.prodhi);			// display frequency
	}
}
